#include "fifocore.h"
#include <windows.h>
#include <sys/cygwin.h>
#include <malloc.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

typedef struct fifoguts
{
  size_t data_offset;
  unsigned int readers,writers;
  size_t buffer_length;
  size_t data_length;
  unsigned int pipeactive;
} fifoguts_t;

typedef struct fifohandle
{
//  HANDLE fileobj;
  HANDLE filemap;
// don't split up the next two variables
  HANDLE readreadyevent;
  HANDLE writingmutex;
  HANDLE writedoneevent;
  HANDLE readmutex;
  HANDLE datamutex;

  void * filedata;
  int read, write;

  char filemapname[MAX_PATH]; /* this may mean that multiple fifos at the max path length with similar prefixes collide. Tough. */
  char rrname[MAX_PATH];
  char rmname[MAX_PATH];
  char wmname[MAX_PATH];
  char wdname[MAX_PATH];
  char rsname[MAX_PATH];
  char wsname[MAX_PATH];
  char dmname[MAX_PATH];
  fifoguts_t *guts;
} fifohandle_t;

void *
openfifo(const char *path, int read, int write)
{
  char win32path[MAX_PATH];
  unsigned int fileaccess=(read ? GENERIC_READ : 0) | (write ? GENERIC_WRITE : 0);
  unsigned int mapaccess;
  unsigned int created=0;
  SYSTEM_INFO sysinfo;
  size_t data_offset;
  fifohandle_t *f=malloc(sizeof(fifohandle_t));
  if (!f)
  {
//    set_errno(ENOMEM);
    return NULL;
  }

//  if (read && write)
    mapaccess=PAGE_READWRITE;
//  else if (read)
//    mapaccess=FILE_MAP_READ;
//  else
//    mapaccess=FILE_MAP_WRITE;
  
  f->read=read;
  f->write=write;  

//  cygwin_conv_to_full_win32_path(path,win32path);
//  printf("%s\n",win32path);  
 
#if 0
  f->fileobj=CreateFile(win32path, 
   fileaccess,0, //no share
  NULL, // no inherit, this username only
  OPEN_EXISTING,
  FILE_ATTRIBUTE_TEMPORARY,0);
  if (f->fileobj==INVALID_HANDLE_VALUE)
  {
    printf("failed to open file\n");
    free(f);
    return NULL;
  }

#endif

// todo enter the FIFO global mutex

// if os=nt4 ts or win2k or above add "Global\" as a prefix. 

// setup our names
  snprintf(f->filemapname,MAX_PATH,"FIFOFM%s",path);
  snprintf(f->rrname,MAX_PATH,"FIFORR%s",path);
  snprintf(f->rmname,MAX_PATH,"FIFORM%s",path);
  snprintf(f->wmname,MAX_PATH,"FIFOWM%s",path);
  snprintf(f->dmname,MAX_PATH,"FIFOWM%s",path);
  snprintf(f->wdname,MAX_PATH,"FIFOWD%s",path);
  snprintf(f->rsname,MAX_PATH,"FIFORS%s",path);
  snprintf(f->wsname,MAX_PATH,"FIFOWS%s",path);

  GetSystemInfo(&sysinfo);
  if (sysinfo.dwAllocationGranularity >=sizeof(fifoguts_t))
    data_offset=sysinfo.dwAllocationGranularity;
  else
    data_offset=sysinfo.dwAllocationGranularity*(1+sizeof(fifoguts_t)/sysinfo.dwAllocationGranularity);

// FIXME: we need two opbjects. One for the header (r/w for every allowed user
// and one for data transfer (access only r or w as per file specs 
  f->filemap=CreateFileMapping(
  INVALID_HANDLE_VALUE,               // system pagefile.
  NULL, // default security: no inherit, this user only FIXME check acl on the file
        // and set this appropriately.
  mapaccess ,                    // protection
  0x00000000,            // high-order DWORD of size (1 Megabyte )+sizeof fifoguts_t
  0x00100000+data_offset,             // low-order DWORD of size
  f->filemapname                      // object name
);

  if (f->filemap==NULL)
  {
    printf("failed to open mapping: %d\n",GetLastError());

  }
  else
  {
  if (GetLastError()!=ERROR_ALREADY_EXISTS)
  created=1;

  f->readreadyevent=CreateEvent(
    NULL, // SD FIXME: check file mask and allow users from that
    FALSE,                       // reset type
    FALSE,                      // initial state
    f->rrname                         // object name
  );

  if (f->readreadyevent==NULL)
  {
    printf("failed to open event : %d\n",GetLastError());
    //close the filemap
  }
  
  f->writingmutex=CreateMutex(
    NULL,  // SD
    NULL,                       // initial owner
    f->wmname                   // object name
  );

  if (f->writingmutex==NULL)
  {
    printf("failed to open write mutex : %d\n",GetLastError());
    //close the event and the filemap
  }

  f->readmutex=CreateMutex(
    NULL,  // SD
    NULL,                       // initial owner
    f->rmname                   // object name
  );

  if (f->readmutex==NULL)
  {
    printf("failed to open read mutex : %d\n",GetLastError());
    //close the event and the filemap  andt eh write mutex
  }

  f->datamutex=CreateMutex(
    NULL,  // SD
    NULL,                       // initial owner
    f->dmname                   // object name
  );

  if (f->datamutex==NULL)
  {
    printf("failed to open read mutex : %d\n",GetLastError());
    //close the event and the filemap  andt eh write mutex
  }
 
  f->writedoneevent=CreateEvent(
    NULL, // SD FIXME: check file mask and allow users from that
    FALSE,                       // reset type
    FALSE,                      // initial state
    f->wdname                         // object name
  );

  if (f->readreadyevent==NULL)
  {
    printf("failed to open event : %d\n",GetLastError());
    //close the event and the filemap  and both mutexs
  }

  f->guts=MapViewOfFile(f->filemap,FILE_MAP_WRITE,0, 0, 0);

  if (f->guts==NULL)
  {
    printf("failed to map fifo file to memory: %d\n",GetLastError());
    //close both event and the filemap  and both mutexs
  }

  // FIXME: SEH for read and writes?

  if (created)
  {
   if (read) f->guts->readers=1; else f->guts->readers=0;
   if (write) f->guts->writers=1; else f->guts->writers=0;
   f->guts->data_offset=data_offset;
   f->guts->buffer_length=0x00100000; //1 Mb
   f->guts->data_length=0;
   f->guts->pipeactive=0; // 0 = pipe waiting for an open
  }
  else
  {
    if (f->guts->data_offset != data_offset)
    {
      printf("incorrect data offset - check for multiple cygwin dll's \n");
      //close both event and the filemap  and both mutexs and free the mapped view
    }
    if (read) f->guts->readers++;
    if (write) f->guts->writers++;
  }

  // FIX ME set the file map read/write to the requested access.
  f->filedata=MapViewOfFile(f->filemap,FILE_MAP_WRITE,0, f->guts->data_offset,
	 f->guts->buffer_length);
  
  if (f->filedata==NULL)
  {
    printf("failed to map fifo data file to memory: %d\n",GetLastError());
    //close both event and the filemap  and both mutexs and the first file view
  }


  }
// todo exit the FIFO global mutex

  return f;
  
}


int fifoclose(void *data)
{
  fifohandle_t *f=(fifohandle_t *)data;
  int rv=0;
  if (!f)
    return EINVAL;
  // FIXME Get the FIFO mutex
  if (!f->filemap)
    rv= EINVAL; // what for already closed files? 
  else 
  {
  if (f->read) f->guts->readers--;
  if (f->write) f->guts->writers--;
  if (f->guts->pipeactive) // at least one i/o has completed
  if (!(f->guts->readers) ||  !(f->guts->writers))
  {
    f->guts->pipeactive=2; // 2 = closed;
    printf("last reader or writer closed the pipe\n");
    printf("Telling the current blocked writers");
    PulseEvent(f->readreadyevent);
    printf("... and readers\n");
    PulseEvent(f->writedoneevent);
  }
  printf("on close: readers %d\n", f->guts->readers);
  printf("on close: writers %d\n", f->guts->writers);
  if (!CloseHandle(f->filemap))rv= -1; /* what error !?!?! */
  if (!CloseHandle(f->readreadyevent))rv= -1; /* what error !?!?! */
  if (!CloseHandle(f->writingmutex))rv= -1; /* what error !?!?! */
  if (!CloseHandle(f->readmutex))rv= -1; /* what error !?!?! */
  if (!CloseHandle(f->datamutex))rv= -1; /* what error !?!?! */
  if (!CloseHandle(f->writedoneevent))rv= -1; /* what error !?!?! */
  if (!UnmapViewOfFile(f->filedata))rv= -1; /* what error !?!?! */
  if (!UnmapViewOfFile(f->guts))rv= -1; /* what error !?!?! */
  }

  free(f); /* error here? */
  // FIXME free the FIFO mutex.
  return rv;
}

size_t fifowrite(void *data, void *buf, size_t length)
{
  fifohandle_t *f=(fifohandle_t *)data;
  unsigned int t;
  if (!f)
    return EINVAL;
  printf("/* wait for the write mutex\n");
  t = WaitForSingleObject(
    f->writingmutex, 
    INFINITE      // time-out interval
  );
  if (f->guts->pipeactive>1)
  {
    // the pipe has been closed
    ReleaseMutex(f->writingmutex);
    //set errno EPIPE
    return 0;
  }
  if (t == WAIT_OBJECT_0 )
  {
    printf("// success \n");
    printf("// are there any readers yet? ");
    t= WaitForSingleObject(f->readmutex,0);
    if (t!=WAIT_TIMEOUT)
    {
      printf(".../* no */\n");
      ReleaseMutex(f->readmutex);
      t = WaitForSingleObject(f->readreadyevent, INFINITE);
      if (t == WAIT_OBJECT_0 )
      {
        // there's a read requested
       if (f->guts->pipeactive>1)
       {
       // the pipe has been closed
         ReleaseMutex(f->writingmutex);
         // set errno EPIPE
         return 0;
        }
      
      } else  { printf("   /* should never happen - write*/ \n");  }
    } else
      printf("yes\n");
     t = WaitForSingleObject(f->datamutex, INFINITE);
    
    // do the write here (we have the write mutex, 
    //and someone else has the read mutex and is waiting
    printf("writing\n");
    // FIXME: multiple writes for big buffers
    memcpy(f->filedata,buf,length);
    f->guts->data_length=length;
    printf("write done, telling reader\n");

    // write done. signal the reader to read and wait for completion 
    
    printf("reader told, waiting for them to accept the data\n");
    ReleaseMutex(f->datamutex);
    t = SignalObjectAndWait(f->writedoneevent, f->datamutex, INFINITE, FALSE);
    if (t == WAIT_OBJECT_0 )
    {
      // they've finished

    } else  {  printf("  /* should never happen */\n");   }

    /* the reason we wait for completion is that another write could overwrite the
     * data space if the reader hasn't copied - and the reader has the mutex so 
     * another writer can't tell
     */
    ReleaseMutex(f->datamutex);
    printf("releasing write mutex\n");
    ReleaseMutex(f->writingmutex);
    return length;
  } else if (t == WAIT_ABANDONED_0 )
  {
    /* success, but the mutex was abandoned. - indicates a previous holder was CTRL-C'd */
    ReleaseMutex(f->writingmutex);
    return fifowrite(data, buf, length);
  }
  /* we don't timeout */
  return length;
}

size_t fiforead(void *data, void *buf, size_t maxlength)
{
  fifohandle_t *f=(fifohandle_t *)data;
  unsigned int t;
  size_t readlength=0;
  if (!f)
    return EINVAL;
  printf("  /* wait for the read mutex */\n");
  t = WaitForSingleObject(
    f->readmutex,        // handle to object
    INFINITE   // time-out interval
  );
  if (f->guts->pipeactive>1)
  {
    printf("// the pipe has been closed\n");
    //seterno EPIPE;
    ReleaseMutex(f->readmutex);
    return 0;
  }
  if (t == WAIT_OBJECT_0)
  {
    printf(" // got the read mutex \n");
    printf(" // are there writers? ...");
    t= WaitForSingleObject(f->writingmutex,0);
    if (t!=WAIT_TIMEOUT)
    {
      printf(" there are no writers. wait for a writer\n");
//    ReleaseMutex(f->writingmutex);
      t = SignalObjectAndWait(f->writingmutex, f->writedoneevent, INFINITE, FALSE);
    }
    else
    {
      printf(" request a write from the waiting writer\n");
//      PulseEvent(f->readreadyevent);
      t = SignalObjectAndWait(f->readreadyevent, f->writedoneevent, INFINITE, FALSE);
    }
    if (f->guts->pipeactive>1)
    {
      printf("// the pipe has been closed\n");
      //set errno EPIPE;
      ReleaseMutex(f->readmutex);
      return 0;
    }

    printf(" wait for a lock on the data \n"); 
    t = WaitForSingleObject(f->datamutex, INFINITE);
    // on infinite rv is irrelevant

    printf("reading");
    // do the read: the write mutex is locked, and we have the read mutex
    // check that the data length <=maxlength. ? do we copy it all out or does the 
    // next reader get the data ?
    memcpy(buf,f->filedata,f->guts->data_length);
    readlength=f->guts->data_length;

    if (!f->guts->pipeactive)
      f->guts->pipeactive=1; // an io has completed
    printf("...signalling acceptance of %d bytes of data...\n",readlength);
    // we've got the data. single complete
//    PulseEvent(f->readreadyevent);
    ReleaseMutex(f->datamutex);

    // read complete
    ReleaseMutex(f->readmutex);
  }
  else if (t == WAIT_ABANDONED)
  {
    //
    printf("abandoned mutex found \n");
  }
  else
  {
   // we don't have the mutex - should never ever happen (no timeout)
  }
//  return 0;
  return readlength;
}
